#!/usr/bin/env python
# -*- coding:utf-8 -*-
# @FileName  :tf_bilstm.py.py
# @Time      :2025/4/14 15:43
# @Author    :David
# @Company: shenyang JY

from tensorflow.keras.layers import Input, Dense, LSTM, concatenate, Conv1D, Conv2D, MaxPooling1D, Reshape, Flatten, Bidirectional, TimeDistributed
from tensorflow.keras.models import Model, load_model
from tensorflow.keras.callbacks import ModelCheckpoint, EarlyStopping, TensorBoard, ReduceLROnPlateau
from tensorflow.keras import optimizers, regularizers
from models_processing.model_tf.losses import region_loss
import numpy as np
from common.database_dml import *
from models_processing.model_tf.settings import set_deterministic
from threading import Lock
import argparse
model_lock = Lock()
set_deterministic(42)

class TSHandler2(object):
    def __init__(self, logger, args):
        self.logger = logger
        self.opt = argparse.Namespace(**args)
        self.model = None
        self.model_params = None

    def get_model(self, args):
        """
        单例模式+线程锁，防止在异步加载时引发线程安全
        """
        try:
            with model_lock:
                loss = region_loss(self.opt)
                self.model, self.model_params = get_keras_model_from_mongo(args, {type(loss).__name__: loss})
        except Exception as e:
            self.logger.info("加载模型权重失败:{}".format(e.args))

    @staticmethod
    def get_keras_model(opt, time_series=3, use_bidirectional=False):
        """
        输入输出对齐策略：
        使用Conv1D的strides=3将48时间步压缩为16（48/3=16）
        通过Bidirectional LSTM进行双向时序特征提取
        用TimeDistributed层实现每个时间步独立预测
        """
        assert time_series >= 3
        loss = region_loss(opt)
        l2_reg = regularizers.l2(opt.Model['lambda_value_2'])
        nwp_input = Input(shape=(opt.Model['time_step']*time_series, opt.Model['input_size']), name='nwp')

        # 时序卷积压缩时间维度
        # 使用kernel_size=3和strides=3将48步压缩到16步
        x = Conv1D(64, kernel_size=time_series, strides=time_series, padding='valid', activation='relu')(nwp_input)  # 输出形状 (16, 64)

        # 双向LSTM捕捉前后时序依赖
        if use_bidirectional:
            x = Bidirectional(LSTM(128, return_sequences=True, kernel_regularizer=l2_reg))(x)  # 输出形状(16, 256)
        else:
            x = LSTM(128, return_sequences=True, kernel_regularizer=l2_reg)(x)  # 保持16时间步
        # 时间分布全连接层
        output = TimeDistributed(Dense(1))(x)  # 输出形状 (16, 1)

        # 压缩最后一个维度
        # output = Dense(opt.Model['time_step']*(time_series-2))(x[:, :, 0:1])  # 最终输出形状 (16,)

        # con1 = Conv1D(filters=64, kernel_size=5, strides=1, padding='valid', activation='relu', kernel_regularizer=l2_reg)(nwp_input)
        # con1_p = MaxPooling1D(pool_size=5, strides=1, padding='valid', data_format='channels_last')(con1)
        # if use_bidirectional:
        #     nwp_lstm = Bidirectional(LSTM(units=opt.Model['hidden_size'], return_sequences=False,  kernel_regularizer=l2_reg), merge_mode='concat')(con1_p) # 默认拼接双向输出（最终维度=2*hidden_size）
        # else:
        #     nwp_lstm = LSTM(units=opt.Model['hidden_size'], return_sequences=False,  kernel_regularizer=l2_reg)(con1_p) # 默认拼接双向输出（最终维度=2*hidden_size）
        #
        # output = Dense(opt.Model['time_step']*(time_series-2), name='cdq_output')(nwp_lstm)

        model = Model(nwp_input, output)
        adam = optimizers.Adam(learning_rate=0.001, beta_1=0.9, beta_2=0.999, epsilon=1e-7, amsgrad=True)
        model.compile(loss=loss, optimizer=adam)
        return model

    def train_init(self):
        try:
            # 进行加强训练，支持修模
            loss = region_loss(self.opt)
            base_train_model, self.model_params = get_keras_model_from_mongo(vars(self.opt), {type(loss).__name__: loss})
            base_train_model.summary()
            self.logger.info("已加载加强训练基础模型")
            return base_train_model
        except Exception as e:
            self.logger.info("加载加强训练模型权重失败:{}".format(e.args))
            return False

    def training(self, model, train_and_valid_data):
        model.summary()
        train_x, train_y, valid_x, valid_y = train_and_valid_data
        early_stop = EarlyStopping(monitor='val_loss', patience=self.opt.Model['patience'], mode='auto')
        history = model.fit(train_x, train_y, batch_size=self.opt.Model['batch_size'], epochs=self.opt.Model['epoch'],
                            verbose=2, validation_data=(valid_x, valid_y), callbacks=[early_stop], shuffle=False)
        loss = np.round(history.history['loss'], decimals=5)
        val_loss = np.round(history.history['val_loss'], decimals=5)
        self.logger.info("-----模型训练经过{}轮迭代-----".format(len(loss)))
        self.logger.info("训练集损失函数为：{}".format(loss))
        self.logger.info("验证集损失函数为：{}".format(val_loss))
        return model

    def predict(self, test_x, batch_size=1):
        result = self.model.predict(test_x, batch_size=batch_size)
        self.logger.info("执行预测方法")
        return result



if __name__ == "__main__":
    run_code = 0